<?php
/*
 * Spigot
 * https://www.spigotmc.org/
 */

class Minecraft_mod_spigot{
	private $server			= array();			# Current server settings loaded
	
	function __construct($server=array()) {
		$this->server = $server;
	}
	
	function option($field, $default=false){
		if(isset($this->server['metadata'][$field])){
			return $this->server['metadata'][$field];
		} else if(isset($this->server[$field])){
			return $this->server[$field];
		}

		return $default;
	}
		
	/* ============================================================
	 * CHAT
	 * ------------------------------------------------------------
	 * Used to parse each line of the log files
	 * Should return either false or a structure array
	 * false in the event of a non-match
	 * structure array depending on properties
	 */

	public $regexp_name		= '[A-z0-9_]{2,16}';
	public $regexp_ip		= '[0-9\.]+';
	public $regexp_coords	= '-?[0-9\.]+,\s*-?[0-9\.]+,\s*-?[0-9\.]+';
	
	private $appendable		= false;
	
	function parse_log_line($line){
		
		
		if(preg_match('/\[([0-9]{1,2}:[0-9]{1,2}:[0-9]{1,2})\]\s+\[([^\]]+)\]:\s+(.*?)$/i', $line, $matches)){
			$this->appendable = false;

			$time	= trim($matches[1]);
			$type	= strtolower(trim($matches[2]));
			$text	= trim($matches[3]);
			
			# Chat
			if(preg_match('/async chat thread/i', $type) && preg_match('/info/i', $type)){
				$this->appendable = true;
				preg_match('/<([^>]+)>\s+(.*?)$/is', $text, $breakdown);
				if($breakdown) {
					return array(
						'time'		=> $time,
						'type'		=> 'chat',
						'user'		=> $breakdown[1],
						'text'		=> $breakdown[2],
					);
				}
			}

			# User Issued Server Command
			# aamche issued server command: /home
			else  if(stripos($text, 'issued server command')!==false){
				preg_match('/(\S+)\s*issued server command:\s+(.*?)$/is', $text, $breakdown);
				
				#echo '<pre>' . print_r($matches, true) . '</pre>';
				#echo '<pre>' . print_r($breakdown, true) . '</pre>';

				$action	= strtolower(trim($breakdown[2]));
			
				// password hide
				if (stripos($action,'/l ')) {				$action = '/l ******'; }
				else if (stripos($action,'/login ')) {		$action = '/login ******'; }
				else if (stripos($action,'/reg ')) {		$action = '/reg ******'; }
				else if (stripos($action,'/register ')) {	$action = '/register ******'; }
				
				return array(
					'time'		=> $time,
					'type'		=> 'command',
					'user'		=> trim($breakdown[1]),
					'text'		=> trim($breakdown[2]),
				);
			}
		}
		
		return false;
	}
	
	
	/* ============================================================
	 * QUERY SERVER
	 * ------------------------------------------------------------
	 * Returns information via directly query to the server
	 */
	
	function query($server=null){
		if($server!==null){$this->server = $server;}
		$this->Connect();
		return $this->GetInfo();
	}
	
	
	/*
	 * Class written by xPaw
	 * Modified for query
	 *
	 * Website: http://xpaw.me
	 * GitHub: https://github.com/xPaw/PHP-Minecraft-Query
	 */

	const STATISTIC = 0x00;
	const HANDSHAKE = 0x09;

	private $Socket;
	private $Players;
	private $Info;

	public function Connect(){
		
		$Ip			= $this->option('host', '127.0.0.1');
		$Port		= $this->option('port', 25565);
		$Timeout	= $this->option('timeout', 10);
		
		$this->query_start = microtime(true);
		
		if( !is_int( $Timeout ) || $Timeout < 0 )
		{
			throw new \InvalidArgumentException( 'Timeout must be an integer.' );
		}
		if (substr_count($Ip , '.') != 4) $Ip = gethostbyname($Ip);

		$this->Socket = @FSockOpen( 'udp://' . $Ip, (int)$Port, $ErrNo, $ErrStr, $Timeout );

		if( $ErrNo || $this->Socket === false )
		{
			log_message( 'Could not create socket: ' . $ErrStr );
		}

		Stream_Set_Timeout( $this->Socket, $Timeout );
		Stream_Set_Blocking( $this->Socket, true );

		try
		{
			$Challenge = $this->GetChallenge( );

			$this->GetStatus( $Challenge );
		}
		// We catch this because we want to close the socket, not very elegant
		catch( MinecraftQueryException $e )
		{
			FClose( $this->Socket );

			log_message( $e->getMessage( ) );
		}

		FClose( $this->Socket );
	}

	public function GetInfo( )
	{
		return isset( $this->Info ) ? $this->Info : false;
	}

	public function GetPlayers( )
	{
		return isset( $this->Players ) ? $this->Players : false;
	}

	private function GetChallenge( )
	{
		$Data = $this->WriteData( self :: HANDSHAKE );

		if( $Data === false )
		{
			log_message( 'Failed to receive challenge.' );
		}

		return Pack( 'N', $Data );
	}

	private function GetStatus( $Challenge )
	{
		$Data = $this->WriteData( self :: STATISTIC, $Challenge . Pack( 'c*', 0x00, 0x00, 0x00, 0x00 ) );

		if( !$Data )
		{
			log_message( 'Failed to receive status.' );
		}

		$Last = '';
		$Info = Array( );

		$Data    = SubStr( $Data, 11 ); // splitnum + 2 int
		$Data    = Explode( "\x00\x00\x01player_\x00\x00", $Data );

		if( Count( $Data ) !== 2 )
		{
			log_message( 'Failed to parse server\'s response.' );
		}

		$Players = SubStr( $Data[ 1 ], 0, -2 );
		$Data    = Explode( "\x00", $Data[ 0 ] );
		$ping = round((microtime(true)-$this->query_start)*1000); //calculate the high five duration

		// Array with known keys in order to validate the result
		// It can happen that server sends custom strings containing bad things (who can know!)
		$Keys = Array(
			'hostname'   => 'motd',
			'gametype'   => 'GameType',
			'version'    => 'Version',
			'plugins'    => 'Plugins',
			'map'        => 'Map',
			'numplayers' => 'Players',
			'maxplayers' => 'MaxPlayers',
			'hostport'   => 'HostPort',
			'hostip'     => 'HostIp',
			'game_id'    => 'GameName'
		);
		
		foreach( $Data as $Key => $Value )
		{
			if( ~$Key & 1 )
			{
				if( !Array_Key_Exists( $Value, $Keys ) )
				{
					$Last = false;
					continue;
				}

				$Last = strtolower($Keys[ $Value ]);
				$Info[ $Last ] = '';
			}
			else if( $Last != false )
			{
				$Info[ $Last ] = mb_convert_encoding( $Value, 'UTF-8' );
			}

		}

		// Ints
		$Info[ 'users' ]    = array(
				'max'		=> (int)$Info[ 'maxplayers' ],
				'online'	=> (int)$Info[ 'players' ],
			);
		if(isset($Players)){
			$Players = explode( "\x00", $Players );
			foreach($Players as $Player){
				$Info[ 'users' ]['sample'][] = array(
						'name'	=> $Player,
					);
			}
		}
		
		$Info[ 'hostport' ]   = IntVal( $Info[ 'hostport' ] );
        $Info[ 'ping' ] = $ping;
		
		if(isset($Info['motd'])){
			$motd = $Info['motd'];
			$motd = preg_replace("/(§.)/", "", $motd);
			$motd = preg_replace("/[^[:alnum:][:punct:] ]/", "", $motd); //Remove all special characters from a string
			$Info['motd'] = $motd;
		}

		// Parse "plugins", if any
		if( $Info[ 'plugins' ] )
		{
			$Data = Explode( ": ", $Info[ 'plugins' ], 2 );

			$Info[ 'rawplugins' ] = $Info[ 'plugins' ];
			$Info[ 'software' ]   = $Data[ 0 ];

			if( Count( $Data ) == 2 )
			{
				$Info[ 'plugins' ] = Explode( "; ", $Data[ 1 ] );
			}
		}
		else
		{
			$Info[ 'software' ] = 'vanilla';
		}

		$this->Info = $Info;

	}

	private function WriteData( $Command, $Append = "" )
	{
		$Command = Pack( 'c*', 0xFE, 0xFD, $Command, 0x01, 0x02, 0x03, 0x04 ) . $Append;
		$Length  = StrLen( $Command );

		if( $Length !== FWrite( $this->Socket, $Command, $Length ) )
		{
			log_message( "Failed to write on socket." );
		}

		$Data = FRead( $this->Socket, 4096 );

		if( $Data === false )
		{
			log_message( "Failed to read from socket." );
		}

		if( StrLen( $Data ) < 5 || $Data[ 0 ] != $Command[ 2 ] )
		{
			return false;
		}

		return SubStr( $Data, 5 );
	}
}
