<?php
/*
 * Main game engine parser file
 * If you are wanting to add additional functionality please add a mod in the 
 * mods folder and call it in config/servers.php
 * 
 * This class calls configured mods to parse files and log
 * lines. This file is just a handler, check out mods/mod_default.php copy it for
 * your own base functionality
 */

class Terraria_exporter {
	private $namespace				= 'terraria';
	private $server					= array();			# Current server settings loaded
	private $mods					= null;	# Mods Objects
	private $mods_parse_log_line	= null;	# Mods with parse log line function
	private $lines_counted			= 0;	# Log row count
	private $lines_parsed			= 0;	# Log rows parsed count
	private $appendable				= null;	# For multilines
	private $log_index				= null;	# Data row for current log open
	private $log_meta				= array();

	/*
	 *	Export method called by exporter/cron
	 */
	function export($server){
		$this->server = $server;
		
		# Load Mods
		$this->mods = (object)array();
		if(		!isset($this->server['mods'])
				|| (is_array($this->server['mods']) && count($this->server['mods'])==0)
				|| (is_string($this->server['mods']) && trim($this->server['mods'])=='')
			){
			$this->server['mods'] = 'default';
		}
		if(!is_array($this->server['mods'])){
			$this->server['mods'] = explode(',', $this->server['mods']);
		}
		if(!in_array('vanilla', $this->server['mods'])){
			$this->server['mods'][] = 'default';
		}
		foreach($this->server['mods'] as $mod){
			$mod = strtolower(trim($mod));
			$filename = BASEPATH . 'extensions/' . $this->namespace . '/mods/mod_' . $mod . '.php';
			if(file_exists($filename)){
				include_once($filename);
				$classname = ucwords($this->namespace) . '_mod_' . $mod;
				$this->mods->$mod = new $classname($server);
				$this->mods->$mod->db = $this->db;	# Load in database
			} else {
				log_message("Missing mod file.\nfile: " . $filename);
			}
		}
		
		# Parse Files
		$this->parse('start');	# Parse before logs are parsed
		$this->parse_logs();	# Parse logs line by line
		$this->parse('end');	# Parse after logs are parsed
	}
	
	/*
	 *	Calls methods in order they are in the config/servers.php mods array
	 *	If vanilla is not called it is called last
	 */
	
	function parse($method_name){
		$method_name = 'export_' . $method_name;
		foreach($this->server['mods'] as $mod){
			$mod = strtolower(trim($mod));
			if(method_exists($this->mods->$mod, $method_name)){
				$this->mods->$mod->$method_name();
			}
		}
	}

	function parse_logs(){
		# Parsers that have parse_log_line method
		foreach($this->server['mods'] as $mod){
			$mod = strtolower(trim($mod));
			if(method_exists($this->mods->$mod, 'parse_log_line')){
				$this->mods_parse_log_line[$mod] = $mod;
			}
		}
		
		if(count($this->mods_parse_log_line)==0){
			return;
		}

		# Begin Main Logs
		$dir 	= rtrim($this->server['directory'], '/');
		$logdir = $dir . '/logs/';
		$logs 	= scandir($logdir);
		sort($logs);

		foreach($logs as $logindex=>$log){
			$logfile = $logdir . $log;
			if(is_file($logfile)){

				# Log name and indexing
				$logfilename = $log;

				# Log index: Select
				$sth = $this->db->conn->prepare("SELECT * FROM log_index WHERE file_name=:file_name AND server_id=:server_id");
				$sth->execute(array(':file_name' => $logfilename, ':server_id' => $this->server['id']));
				$this->log_index = $sth->fetch(PDO::FETCH_ASSOC);
				
				# If log is already complete skip to next one
				if(isset($this->log_index['status']) && $this->log_index['status']=='complete'){
					continue;
				}
				
				# Is Complete Log
				if($logindex>=(count($logs)-2)){
					$complete_file = false;
				} else {
					$complete_file = true;
				}
				
				# Clear out this log
				if(isset($this->log_index['log_id'])){
					$this->db->delete('log_activity', array(
								'server_id'		=> $this->server['id'],
								'log_id'		=> $this->log_index['log_id'],
							)
						);
				} else {
					$this->log_index['server_id'] 		= $this->server['id'];
					$this->log_index['datetime_file'] 	= date('Y-m-d', $this->log_meta['datetime_unix']);
					$this->log_index['file_name'] 		= $logfilename;
					$this->db->insert('log_index', $this->log_index);
					$this->log_index['log_id'] 			= $this->db->conn->lastInsertId();
				}

				# Process Log
				$this->lines_counted = 0;	# Line Count
				$this->lines_parsed = 0;	# Added Line Count
				
				# Gzip File
				if(substr($log, -3)=='.gz'){
					$handle = gzopen($logfile, 'r');
					while (!gzeof($handle)) {
						$line = gzgets($handle, 4096);
						$this->line($line);
					}
					gzclose($handle);

				}
				# Open File
				else {
					$handle = fopen($logfile, "r");
					if ($handle) {
					
						$this->appendable = null;
						while (($line = fgets($handle)) !== false) {
							$this->line($line);
						}
					} else {
						$data['errors'][] = 'Unabled to open: ' . $logfile;
					} 
					fclose($handle);
				}
					
				// Catch any last appendable rows
				if($this->appendable!==null){
					$this->lines_counted++;
					$this->lines_parsed++;
					$this->store->set($this->appendable);
					$this->appendable = null;
				}

				# Log Index: Update
				$this->db->update('log_index', array(
							'datetime_parsed'	=> date('Y-m-d H:i:s'),
							'status'			=> ($complete_file===true ? 'complete' : null),
							'file_md5'			=> md5_file($logfile),
							'lines_parsed'		=> $this->lines_parsed,
							'lines_counted'		=> $this->lines_counted,
						), array(
							'server_id'			=> $this->server['id'],
							'log_id'			=> $this->log_index['log_id'],
						)
					);
			}
		}
		
	}
	
	function line($line){
		# Parse Line, loop through mods
		# parse_log_line should return an array or false
		$this->lines_counted++;
		foreach($this->mods_parse_log_line as $mod){
			$result = $this->mods->$mod->parse_log_line($line);
			if($result!==false){
				break;
			}
		}

		# If a row we can process
		if(isset($result['type'])){
			$this->lines_parsed++;
			#echo '<pre>' . print_r($result, true) . '</pre>';
			echo '. ';
			flush();
			
			$db = $result;
			$db['server_id']	= $this->server['id'];
			$db['log_id']		= $this->log_index['log_id'];
			if(isset($db['time'])){
				$db['datetime']		= date('Y-m-d H:i:s', strtotime($db['time']));
			}

			# Appendable: This row could have more data added
			if(isset($result['appendable']) && $result['appendable']===true){
				if($this->appendable!==null){
					$this->store->set($this->appendable);
				}
				$this->appendable = $db;
			} 
			# Append: to previous rows
			else if($result['type']==='append' && $this->appendable!==null){
				$this->appendable['text'] .= "\r\n" . $result['text'];
			}
			# Store Row
			else {
				if($this->appendable!==null){
					$this->store->set($this->appendable);
					$this->appendable = null;
				}
				$this->store->set($db);
			}
		}
	}
}
